
/*
 * LFN.C
 *
 * This library contains replacement file handling functions which support
 * long filenames in DOS boxes under Windows 95/98.
 *
 * This library was developed using MS-DOS but should also be portable to
 * WIN16, WIN32 and UNIX environments.
 *
 * Freeware by Paul Burgin (m6809@usa.net).
 * 28/2/98.
 */

#if !defined(LFN95) && !defined(WIN32)
#if defined(__MSDOS__) || defined(WIN16)
#define LFN95
#endif
#endif

#include <stdio.h>		/* fopen, freopen, _fsopen, remove, rename */
#include <sys/stat.h>	/* stat, mkdir, chmod */
#if defined(LFN95) || defined(WIN32)
#include <io.h>			/* access, chmod */
#include <dir.h>		/* MAXPATH, chdir, mkdir, rmdir */
#else
#include <unistd.h>		/* access, chdir, rmdir */
#endif
#if defined(LFN95)
#include <string.h>		/* strchr */
#include <dos.h>		/* struct REGPACK, intr, FP_SEG, FP_OFF, _dos_close */

#define CFLAG	1

enum { eInit95, eNo95, eYes95 } eWin95;

/*********************
 * Private Functions *
 ********************/

/*
 * Find short equivalent to long filename by calling int 21h function
 * 7160h.
 */
static int bFindShortName(struct REGPACK *pstRegs, char *szShort, const char *szLong)
{
	pstRegs->r_ax = 0x7160;
	pstRegs->r_cx = 0x0001;
	pstRegs->r_ds = FP_SEG((void far *)szLong);
	pstRegs->r_si = FP_OFF((void far *)szLong);
	pstRegs->r_es = FP_SEG((void far *)szShort);
	pstRegs->r_di = FP_OFF((void far *)szShort);
	intr(0x21, pstRegs);
	return((pstRegs->r_flags & CFLAG) == 0);
}

/*
 * Create an empty file with the specified long filename by calling
 * int 21h function 716ch.
 */
static int bTouchLongName(struct REGPACK *pstRegs, const char *szLong)
{
	pstRegs->r_ax = 0x716c;
	pstRegs->r_bx = 0x2001;
	pstRegs->r_cx = 0x0000;
	pstRegs->r_dx = 0x0012;
	pstRegs->r_ds = FP_SEG((void far *)szLong);
	pstRegs->r_si = FP_OFF((void far *)szLong);
	intr(0x21, pstRegs);
	if ((pstRegs->r_flags & CFLAG) == 0)
	{
		return(_dos_close(pstRegs->r_ax) == 0);
	}
	return(0);
}

/*
 * Detect Windows 95/98 by calling int 21h function 3000h to get MS-DOS
 * version, then int 2fh function 4a33h to check for Windows 95/98,
 * then call a Windows 95/98 service and check for success.
 */
static int bWin95(struct REGPACK *pstRegs, char *szShort)
{
	if (eWin95 == eInit95)
	{
		eWin95 = eNo95;
		pstRegs->r_ax = 0x3000;
		intr(0x21, pstRegs);
		if ((pstRegs->r_ax & 0x00ff) >= 7)
		{
			pstRegs->r_ax = 0x4a33;
			intr(0x2f, pstRegs);
			if (pstRegs->r_ax == 0)
			{
				if (bFindShortName(pstRegs, szShort, "."))
				{
					eWin95 = eYes95;
				}
			}
		}
	}
	return(eWin95 == eYes95);
}
#endif

/********************
 * Public Functions *
 ********************/

/* lfn_is95() returns 1 if running under Windows 95/98. */
int lfn_is95(void)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortDir[MAXPATH];

	return(bWin95(&stRegs, szShortDir));
#else
	return(0);
#endif
}

/* lfn_fopen() is a direct replacement for fopen(). */
FILE *lfn_fopen(const char *szLongName, const char *szMode)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortName[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortName))
	{
		if (bFindShortName(&stRegs, szShortName, szLongName))
		{
			/* Found file with specified name - so open it. */
			return(fopen(szShortName, szMode));
		}
		else if (strchr(szMode, 'w'))
		{
			/* File doesn't exist - so create it if opening for write. */
			if (bTouchLongName(&stRegs, szLongName))
			{
				if (bFindShortName(&stRegs, szShortName, szLongName))
				{
					/* Open newly created file. */
					return(fopen(szShortName, szMode));
				}
			}
			if (stRegs.r_ax != 0x7100)
			{
				/* Create failed. */
				return(NULL);
			}
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(fopen(szLongName, szMode));
}

/* lfn_freopen() is a direct replacement for freopen(). */
FILE *lfn_freopen(const char *szLongName, const char *szMode, FILE *pstStream)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortName[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortName))
	{
		if (bFindShortName(&stRegs, szShortName, szLongName))
		{
			/* Found file with specified name - so open it. */
			return(freopen(szShortName, szMode, pstStream));
		}
		else if (strchr(szMode, 'w'))
		{
			/* File doesn't exist - so create it if opening for write. */
			if (bTouchLongName(&stRegs, szLongName))
			{
				if (bFindShortName(&stRegs, szShortName, szLongName))
				{
					/* Open newly created file. */
					return(freopen(szShortName, szMode, pstStream));
				}
			}
			if (stRegs.r_ax != 0x7100)
			{
				/* Create failed. */
				return(NULL);
			}
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(freopen(szLongName, szMode, pstStream));
}

#if defined(LFN95) || defined(WIN32)
/* _lfn_fsopen() is a direct replacement for _fsopen(). */
FILE *_lfn_fsopen(const char *szLongName, const char *szMode, int iShFlg)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortName[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortName))
	{
		if (bFindShortName(&stRegs, szShortName, szLongName))
		{
			/* Found file with specified name - so open it. */
			return(_fsopen(szShortName, szMode, iShFlg));
		}
		else if (strchr(szMode, 'w'))
		{
			/* File doesn't exist - so create it if opening for write. */
			if (bTouchLongName(&stRegs, szLongName))
			{
				if (bFindShortName(&stRegs, szShortName, szLongName))
				{
					/* Open newly created file. */
					return(_fsopen(szShortName, szMode, iShFlg));
				}
			}
			if (stRegs.r_ax != 0x7100)
			{
				/* Create failed. */
				return(NULL);
			}
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(_fsopen(szLongName, szMode, iShFlg));
}
#endif

/* lfn_access() is a direct replacement for access(). */
int lfn_access(const char *szLongName, int iMode)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortName[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortName))
	{
		if (bFindShortName(&stRegs, szShortName, szLongName))
		{
			/* Found file with specified name - so access it. */
			return(access(szShortName, iMode));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(access(szLongName, iMode));
}

/* lfn_chdir() is a direct replacement for chdir(). */
int lfn_chdir(const char *szLongPath)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		if (bFindShortName(&stRegs, szShortPath, szLongPath))
		{
			/* Found specified directory - so enter it. */
			return(chdir(szShortPath));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(chdir(szLongPath));
}

/* lfn_stat() is a direct replacement for stat(). */
int lfn_stat(char *szLongPath, /* struct stat */ void *szStatBuf)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		if (bFindShortName(&stRegs, szShortPath, szLongPath))
		{
			/* Found specified file or directory - get stats for it. */
			return(stat(szShortPath, szStatBuf));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(stat(szLongPath, szStatBuf));
}

/* lfn_remove() is a direct replacement for remove(). */
int lfn_remove(const char *szLongName)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortName[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortName))
	{
		if (bFindShortName(&stRegs, szShortName, szLongName))
		{
			/* Found file with specified name - so remove it. */
			return(remove(szShortName));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(remove(szLongName));
}

/* lfn_mkdir() is a direct replacement for mkdir(). */
#if defined(LFN95) || defined(WIN32)
int lfn_mkdir(const char *szLongPath)
#else
int lfn_mkdir(const char *szLongPath, int iMode)
#endif
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		/*
		 * Create an directory with the specified long filename by calling
		 * int 21h function 7139h.
		 */
		stRegs.r_ax = 0x7139;
		stRegs.r_ds = FP_SEG((void far *)szLongPath);
		stRegs.r_dx = FP_OFF((void far *)szLongPath);
		intr(0x21, &stRegs);
		if ((stRegs.r_flags & CFLAG) == 0)
		{
			return(0);
		}
		if (stRegs.r_ax != 0x7100)
		{
			return(-1);
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
#if defined(LFN95) || defined(WIN32)
	return(mkdir(szLongPath));
#else
	return(mkdir(szLongPath, iMode));
#endif
}

/* lfn_rmdir() is a direct replacement for rmdir(). */
int lfn_rmdir(const char *szLongPath)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		if (bFindShortName(&stRegs, szShortPath, szLongPath))
		{
			/* Found specified directory - so delete it. */
			return(rmdir(szShortPath));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(rmdir(szLongPath));
}

/* lfn_chmod() is a direct replacement for chmod(). */
int lfn_chmod(const char *szLongPath, int iMode)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		if (bFindShortName(&stRegs, szShortPath, szLongPath))
		{
			/* Found specified directory - so chmod it. */
			return(chmod(szShortPath, iMode));
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(chmod(szLongPath, iMode));
}

/* lfn_rename() is a direct replacement for rename(). */
int lfn_rename(const char *szOldName, const char *szNewName)
{
#if defined(LFN95)
	struct REGPACK	stRegs;
	char			szShortPath[MAXPATH];

	/* Use Windows 95/98 long filename services if available. */
	if (bWin95(&stRegs, szShortPath))
	{
		/*
		 * Rename file or directory with long filenames by calling
		 * int 21h function 7156h.
		 */
		stRegs.r_ax = 0x7156;
		stRegs.r_ds = FP_SEG((void far *)szOldName);
		stRegs.r_dx = FP_OFF((void far *)szOldName);
		stRegs.r_es = FP_SEG((void far *)szNewName);
		stRegs.r_di = FP_OFF((void far *)szNewName);
		intr(0x21, &stRegs);
		if ((stRegs.r_flags & CFLAG) == 0)
		{
			return(0);
		}
		if (stRegs.r_ax != 0x7100)
		{
			return(-1);
		}
	}
#endif
	/* If Windows 95/98 not available then pass request to C runtime. */
	return(rename(szOldName, szNewName));
}
